def join_path(row_path, column_path):
    if column_path is None:
        raise Exception("Column path is empty!")
    if row_path is None:
        raise Exception("Row path is empty")
    
    column_path = column_path.strip()
    row_path = row_path.strip()
    
    if len(row_path) > 0 and row_path[-1] == ".":
        row_path = row_path[0:-1]
    if len(column_path) > 0 and column_path[-1] == ".":
        column_path = column_path[0:-1]
    if len(row_path) > 0 and row_path[0] == ".":
        row_path = row_path[1:]
    if len(column_path) and column_path[0] != ".":
        column_path = "." + column_path
        
    if row_path == "":
        return column_path[1:] if len(column_path) > 1 else "."
    if column_path == ".":
        return row_path

    return row_path+column_path


from .set_in import set_in


def flatten(orig_dict, path_prefix=""):
    flat_dict = {}

    if isinstance(orig_dict, list):
        for index, item in enumerate(orig_dict):
            if item is None:
                continue
            if isinstance(item, list) or isinstance(item, dict):
                flat_dict.update(flatten(item, path_prefix + str(index) + "."))
            else:
                flat_dict[path_prefix + str(index)] = item
    elif isinstance(orig_dict, dict):
        for key, value in orig_dict.items():
            if value is None:
                continue
            if isinstance(value, list) or isinstance(value, dict):
                flat_dict.update(flatten(value, path_prefix + key + "."))
            else:
                flat_dict[path_prefix + key] = value
    else:
        raise ValueError("Invalid Type for flatten, needs to be list or dict")

    return flat_dict


def unflatten(flattened_dict):
    nested_dict = {}

    for key, value in flattened_dict.items():
        nested_dict = set_in(nested_dict, key, value)

    return nested_dict


if __name__ == "__main__":
    assert flatten({"a": 1})["a"] == 1
    assert flatten({"a": [None, 2]})["a.1"] == 2
    assert flatten({"a": [None, {"b": 3}]})["a.1.b"] == 3

    assert unflatten({"a": 1}) == {"a": 1}
    assert unflatten({"a.1": 2}) == {"a": [None, 2]}
    assert unflatten({"a.1.b": 3}) == {"a": [None, {"b": 3}]}



import re

# paths          -> return_val
# ----------------------------
# interface      -> interface
# samples.*      -> samples.1
# samples.*.item -> samples.1.item
# samples.*      -> samples.2
# metadata       -> metadata

def should_star_increment(path, prefix):
    # Remove initial part of string to see if there's anything at the end e.g.
    # samples.* -> True
    # samples.*.item1 -> False
    
    remaining_string = path.replace(prefix, "")
    remaining_string = re.sub("\*\]?", "", remaining_string)
    return len(remaining_string) == 0

def replace_path_stars(paths):
    new_paths = []

    paths_with_index = enumerate(paths)
    for (path_index, path) in paths_with_index:
        
        if not "*" in path:
            new_paths.append(path)
            continue
        
        prefix = re.sub('\[?\*.*', "", path)
        
        last_number = None
        for already_added_path in new_paths[::-1]:
            if already_added_path.startswith(prefix):
                already_added_path_wo_prefix = already_added_path.replace(prefix, "")
                get_num_re = re.compile("\[?([0-9]+)")
                last_number = int(get_num_re.search(already_added_path_wo_prefix).group(1))
                break
        
        new_path = None
        if last_number is None:
            new_path = path.replace("*", "0")
        elif should_star_increment(path, prefix):
            new_path = path.replace("*", str(last_number + 1))
        else:
            new_path = path.replace("*", str(last_number))
    
        new_paths.append(new_path)
        
    return new_paths


from jac_format.join_path import join_path
from jac_format.replace_path_stars import replace_path_stars
from jac_format.to_json import to_json
from jac_format.to_csv import to_csv

from_json = to_csv
from_csv = to_json



from .join_path import join_path
from .get_in import get_in
from .set_in import set_in
from .flatten import flatten, unflatten


def remove_redundancies(rows=None, columns=None, array=None):
    if rows is None:
        raise ValueError("rows must be specified in remove_redundancy")
    if columns is None:
        raise ValueError("columns must be specified in remove_redundancy")
    if array is None:
        raise ValueError("array must be specified in remove_redundancy")

    all_paths = []
    for ri, row in enumerate(rows):
        for ci, col in enumerate(columns):
            all_paths.append((join_path(row, col), ri + 1, ci + 1))

    all_paths = all_paths[::-1]

    reconstructed_object = {}
    for base_path, rowi, coli in all_paths:
        cell_value = array[rowi][coli]
        if isinstance(cell_value, dict):
            flattened_obj = flatten(array[rowi][coli])
            paths_to_remove = {}
            for sub_path in flattened_obj.keys():
                full_path = join_path(base_path, sub_path)
                existing_val = get_in(reconstructed_object, full_path)
                if (
                    existing_val is not None
                    and flattened_obj.get(sub_path, None) is None
                ):
                    paths_to_remove[sub_path] = True
                    continue
                if existing_val == flattened_obj[sub_path]:
                    paths_to_remove[sub_path] = True
                    continue
            if bool(paths_to_remove):
                new_cell_value = {}
                for sub_path in flattened_obj:
                    if sub_path not in paths_to_remove:
                        new_cell_value = set_in(
                            new_cell_value, sub_path, flattened_obj[sub_path]
                        )
                if not bool(new_cell_value):
                    array[rowi][coli] = None
                else:
                    array[rowi][coli] = new_cell_value
                    cell_value = new_cell_value
        else:  # cell value is not dict
            existing_value = get_in(reconstructed_object, base_path)
            if existing_value == cell_value:
                array[rowi][coli] = None

        cell_value = array[rowi][coli]

        # Update reconstructed object
        if isinstance(cell_value, dict) and bool(
            get_in(reconstructed_object, base_path)
        ):
            reconstructed_object = set_in(
                reconstructed_object,
                base_path,
                unflatten(
                    flatten(cell_value),
                    flatten(get_in(reconstructed_object, base_path)),
                ),
            )
        elif cell_value is not None:
            reconstructed_object = set_in(reconstructed_object, base_path, cell_value)
    return array



import csv
from functools import reduce
from .replace_path_stars import replace_path_stars
from .join_path import join_path
from .get_cell_value import get_cell_value
from .set_in import set_in


def to_json(csv_string, options={}):
    reader = csv.reader(csv_string.split("\n"), delimiter=",")
    rows = []
    for row in reader:
        rows.append(row)

    print("rows", rows)

    if rows[0][0] != "path" and rows[0][0] != "jac_csv_path":
        if not "derivePath" in options:
            raise Exception(
                'No "path" or "jac_csv_path" in first cell (make sure this file is formatted in the JAC format https://github.com/seveibar/jac-format)'
            )
        # rows_with_index = enumerate(rows)
        # for row_with_index in rows_with_index:
        #     row_index = row_with_index[0]
        #     row_object = rows[row_index]
        #     row_object = reduce(lambda x, y:)
        #     # line 16

        #     # new_paths =

    new_paths = replace_path_stars([row[0] for row in rows[1:]])

    for row_index in range(1, len(rows)):
        rows[row_index][0] = new_paths[row_index - 1]

    header = replace_path_stars(
        [row if row[0] == "." else "." + row for row in rows[0]]
    )

    obj = {}
    for row in rows[1:]:
        for cell_index in range(1, len(row)):
            if header[cell_index] is None:
                continue
            if row[0] is None:
                continue

            full_path = join_path(row[0], header[cell_index])
            cell_value = get_cell_value(row[cell_index])

            if cell_value is None:
                continue

            if full_path == ".":
                obj = cell_value
                continue
            obj = set_in(obj, full_path, cell_value)

    return obj



import csv
import io
from .join_path import join_path
from .get_in import get_in
from .remove_redundancies import remove_redundancies
from sys import version_info


def to_csv(
    json, rows=["."], columns=["."], validate=True, should_remove_redundancies=True
):

    ar = [["path"] + columns]

    # Normalize column definitions
    columns = [c if c.startswith(".") else ".{}".format(c) for c in columns]

    # If the column refers to an array, note the path to the item's parent
    column_parent_array_path = []
    for c in columns:
        try:
            int(c.split(".")[-1])
            column_parent_array_path.append(".".join(c.split(".")[:-1]))
        except:
            column_parent_array_path.append(None)

    for rowi, row in enumerate(rows):
        new_row = [row]
        for coli, col in enumerate(columns):
            full_path = join_path(row, col)
            if full_path == ".":
                new_row.append(json)
                continue

            if column_parent_array_path[coli] is not None:
                # make sure the part before the index is not a string
                # i.e. the user can't use "string.0" as a path to the first char
                if isinstance(
                    get_in(json, join_path(row, column_parent_array_path[coli])), str
                ):
                    new_row.append(None)
                    continue

            new_row.append(get_in(json, full_path))
        ar.append(new_row)

    # ar is now in 2d redundant jac-format

    if should_remove_redundancies:
        remove_redundancies(rows=rows, columns=columns, array=ar)

    # make sure each cell value is in unicode
    for rowi in range(len(ar)):
        for coli in range(len(ar[rowi])):
            if ar[rowi][coli] is None:
                ar[rowi][coli] = ""
            else:
                ar[rowi][coli] = str(ar[rowi][coli])

    # turn ar into CSV
    if version_info.major == 2:
        output = io.BytesIO()
    else:
        output = io.StringIO()
    writer = csv.writer(output)
    writer.writerows(ar)
    result = output.getvalue()

    return result



import re


def set_in_path_array(d, path, val, prev_path=[]):
    if len(path) == 0:
        return val

    seg = path[0]

    if seg.isdigit() and not (isinstance(d, list) or d is None):
        raise ValueError(
            "Path didn't lead to list, but was being indexed like a list {}".format(
                ".".join(prev_path)
            )
        )
    if seg.isdigit():
        seg = int(seg)
        if d is None:
            d = []

    if d is None:
        d = {}

    if len(path) == 1:
        if isinstance(seg, int):
            while len(d) <= seg:
                d.append(None)
            d[seg] = val
            return d
        else:
            d[seg] = val
            return d
    else:  # len(path) > 1
        if isinstance(d, list):
            while len(d) <= seg:
                d.append(None)
            d[seg] = set_in_path_array(d[seg], path[1:], val, prev_path + [seg])
        else:
            d[seg] = set_in_path_array(
                d.get(seg, None), path[1:], val, prev_path + [seg]
            )
        return d


def set_in(d, path, val):
    path = re.sub(r"\[([0-9]+)\]", r".\1", path)
    return set_in_path_array(d, path.split("."), val)


if __name__ == "__main__":
    a = {}
    a = set_in(a, "b.c", "d")
    assert a["b"]["c"] == "d"

    a = []
    a = set_in(a, "0.b.1", "c")
    assert a[0]["b"][1] == "c"

    a = {}
    a = set_in(a, "b[1].c", "d")
    assert a["b"][1]["c"] == "d"



from os.path import join, abspath, dirname
from setuptools import setup, find_packages

_here = abspath(dirname(__file__))


with open(join(_here, "../README.md")) as f:
    readme = f.read()

with open(join(_here, "../LICENSE")) as f:
    license = f.read().replace("\n", "")

setup(
    name="jac_format",
    version="0.1.1",
    description="JAC (JSON as CSV) Format Conversion",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="UniversalDataTool",
    author_email="seve@wao.ai",
    url="https://github.com/UniversalDataTool/jac-format",
    license=license,
    packages=["jac_format"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
    ],
)



import re


def get_in_array_path(d, path):
    if d is None:
        return None
    if len(path) == 0:
        return d
    if isinstance(d, list):
        if not isinstance(path[0], int):
            return None
        if path[0] >= len(d):
            return None
        return get_in_array_path(d[path[0]], path[1:])
    else:
        return get_in_array_path(d.get(path[0], None), path[1:])


def get_in(d, path):
    path = re.sub(r"\[([0-9]+)\]", r".\1", path)
    path = [int(seg) if seg.isdigit() else seg for seg in path.split(".")]
    return get_in_array_path(d, path)


if __name__ == "__main__":
    assert get_in([{"b": [None, "c"]}], "0.b.1") == "c"
    assert a[0]["b"][1] == "c"
